INTEGER, PLUS, MINUS, MUL, DIV, EOF, LPAREN, RPAREN = 'INTEGER', 'PLUS', 'MINUS', 'MUL', 'DIV', 'EOF', 'LPAREN', 'RPAREN'  # 新增加左右括号常量


class Token:
    def __init__(self, value_type, value):
        self.value_type = value_type
        self.value = value

    def __str__(self):
        return 'Token({value_type},{value})'.format(value_type=self.value_type, value=self.value)

    def __repr__(self):
        return self.__str__()


class Lexer:
    def __init__(self, text):
        self.text = text
        self.position = 0
        self.current_char = self.text[self.position]

    def error(self):
        raise Exception('警告：错误的输入内容！')

    def skip_whitespace(self):
        while self.current_char is not None and self.current_char.isspace():
            self.advance()

    def long_integer(self):
        result = ''
        while self.current_char is not None and self.current_char.isdigit():
            result += self.current_char
            self.advance()
        return int(result)

    def advance(self):
        self.position += 1
        if self.position >= len(self.text):
            self.current_char = None
        else:
            self.current_char = self.text[self.position]

    def get_next_token(self):
        while self.current_char is not None:
            if self.current_char.isspace():
                self.skip_whitespace()
                continue
            if self.current_char.isdigit():
                return Token(INTEGER, self.long_integer())
            if self.current_char == '+':
                self.advance()
                return Token(PLUS, '+')
            if self.current_char == '-':
                self.advance()
                return Token(MINUS, '-')
            if self.current_char == '*':
                self.advance()
                return Token(MUL, '*')
            if self.current_char == '/':
                self.advance()
                return Token(DIV, '/')
            if self.current_char == '(':
                self.advance()
                return Token(LPAREN, '(')
            if self.current_char == ')':
                self.advance()
                return Token(RPAREN, ')')
            self.error()
        return Token(EOF, None)


class AST:  # 定义抽象语法树类作为基类
    pass  # 暂时无内容


class BinaryOperator(AST):  # 定义二元运算符节点类
    def __init__(self, left, operator, right):  # 初始化
        self.left = left  # 运算符左侧子节点
        self.token = self.operator = operator  # 运算符记号和运算符节点
        self.right = right  # 运算符右侧子节点


class Num(AST):  # 定义数字节点类
    def __init__(self, token):  # 初始化
        self.token = token  # 数字记号
        self.value = token.value  # 数字的值


class UnaryOperator(AST):  # 定义一元运算符节点类
    def __init__(self, operator, expr):  # 初始化
        self.token = self.operator = operator  # 运算符记号和运算符节点
        self.expr = expr  # 一元运算符的子表达式


class Parser:  # 定义语法分析器的类
    def __init__(self, lexer):
        self.lexer = lexer  # 接收词法分析器对象
        self.current_token = lexer.get_next_token()

    def error(self):  # 增加语法分析器的错误方法
        raise Exception('警告：错误的输入内容！')

    def eat(self, value_type):
        if self.current_token.value_type == value_type:
            self.current_token = self.lexer.get_next_token()
        else:
            self.error()

    def factor(self):
        current_token = self.current_token
        if current_token.value_type == PLUS:  # 如果是加法运算符
            self.eat(PLUS)
            node = UnaryOperator(current_token, self.factor())  # 运算符记号和子表达式组成节点对象
            return node
        elif current_token.value_type == MINUS:  # 如果是减法运算符
            self.eat(MINUS)
            node = UnaryOperator(current_token, self.factor())  # 运算符记号和子表达式组成节点对象
            return node
        elif current_token.value_type == INTEGER:
            self.eat(INTEGER)
            return Num(current_token)
        elif current_token.value_type == LPAREN:
            self.eat(LPAREN)
            node = self.expr()
            self.eat(RPAREN)
            return node

    def term(self):
        node = self.factor()  # 左侧节点对象
        while self.current_token.value_type in (MUL, DIV):
            token = self.current_token
            if token.value_type == MUL:
                self.eat(MUL)
            elif token.value_type == DIV:
                self.eat(DIV)
            node = BinaryOperator(node, token, self.factor())  # 创建运算符节点对象
        return node  # 返回节点对象

    def expr(self):
        node = self.term()  # 左侧节点对象
        while self.current_token.value_type in (PLUS, MINUS):
            token = self.current_token
            if token.value_type == PLUS:
                self.eat(PLUS)
            elif token.value_type == MINUS:
                self.eat(MINUS)
            node = BinaryOperator(node, token, self.term())  # 返回二元运算表达式的节点
        return node  # 返回运算符节点对象

    def parser(self):  # 定义语法分析器的方法
        return self.expr()  # 返回树对象


class NodeVisitor:  # 定义节点访问器的类
    def visit(self, node):  # 定义访问方法
        method_name = 'visit_' + type(node).__name__  # 获取节点类型名称组成访问器方法名（子类Interpreter中方法的名称）
        visitor = getattr(self, method_name, self.generic_visit)  # 获取访问器对象，找不到访问器时获取“generic_visit”方法对象。
        return visitor(node)  # 返回访问器访问节点的结果

    def generic_visit(self, node):  # 定义通用访问方法
        raise Exception(f'未找到“visit_{type(node).__name__}()”方法！')  # 抛出异常


class Interpreter(NodeVisitor):  # 定义解释器类（继承节点访问器类）
    def __init__(self, parser):  # 初始化方法interpreter07.py
        self.parser = parser  # 接收语法分析器对象

    def visit_BinaryOperator(self, node):  # 访问二元运算符类型节点的方法
        if node.operator.value_type == PLUS:  # 如果操作符类型是加法
            return self.visit(node.left) + self.visit(node.right)  # 分别访问左侧和右侧的记号，将获取的值进行加法运算，并返回结果。
        elif node.operator.value_type == MINUS:
            return self.visit(node.left) - self.visit(node.right)
        elif node.operator.value_type == MUL:
            return self.visit(node.left) * self.visit(node.right)
        elif node.operator.value_type == DIV:
            return self.visit(node.left) // self.visit(node.right)

    def visit_Num(self, node):  # 访问数字类型节点的方法
        return node.value

    def visit_UnaryOperator(self, node):  # 访问一元运算符类型节点的方法
        if node.operator.value_type == PLUS:  # 如果运算符类型是加号
            return +self.visit(node.expr)  # 返回正的访问子节点的结果
        if node.operator.value_type == MINUS:  # 如果运算符类型是减号
            return -self.visit(node.expr)  # 返回负的访问子节点的结果

    def interpret(self):  # 执行解释的方法
        tree = self.parser.parser()  # 获取语法分析器分析后的树对象
        return self.visit(tree)  # 返回访问树对象的计算结果


def main():
    while True:
        try:
            text = input('SPI>>>')  # 改个样式，看着更专业一些：）
        except EOFError:
            break
        if not text:
            continue
        lexer = Lexer(text)
        parser = Parser(lexer)  # 传入词法分析器对象实例化语法分析器
        interpreter = Interpreter(parser)  # 传入语法分析器对象实例化解释器
        result = interpreter.interpret()  # 解释器对象调用解释方法获取结果
        print(text, '=', result)


if __name__ == '__main__':
    main()
